@@ -29,6 +29,14 @@ Q_DECLARE_METATYPE(Plasma::DataEngine::Data)
 #include "serviceviewer.h"
 #include "titlecombobox.h"
 
+enum {
+    ColumnDataSourceAndKey = 0,
+    ColumnType = 1,
+    ColumnValue = 2,
+
+    ColumnCount,
+};
+
 EngineExplorer::EngineExplorer(QWidget *parent)
     : QDialog(parent)
     , m_engine(nullptr)
@@ -118,6 +126,13 @@ void EngineExplorer::setInterval(const int interval)
     m_updateInterval->setValue(interval);
 }
 
+void EngineExplorer::removeExtraRows(QStandardItem *parent, int preserve)
+{
+    if (parent->rowCount() > preserve) {
+        parent->removeRows(preserve, parent->rowCount() - preserve);
+    }
+}
+
 void EngineExplorer::dataUpdated(const QString &source, const Plasma::DataEngine::Data &data)
 {
     QList<QStandardItem *> items = m_dataModel->findItems(source, Qt::MatchExactly);
@@ -129,10 +144,7 @@ void EngineExplorer::dataUpdated(const QString &source, const Plasma::DataEngine
     QStandardItem *parent = items.first();
 
     int rows = showData(parent, data);
-
-    while (parent->rowCount() > rows) {
-        parent->removeRow(rows);
-    }
+    removeExtraRows(parent, rows);
 }
 
 void EngineExplorer::listEngines()
@@ -164,10 +176,10 @@ void EngineExplorer::showEngine(const QString &name)
     m_serviceRequesterButton->setEnabled(false);
     enableButtons(false);
     m_dataModel->clear();
-    m_dataModel->setColumnCount(4);
-    QStringList headers;
-    headers << i18n("DataSource") << i18n("Key") << i18n("Value") << i18n("Type");
+    m_dataModel->setColumnCount(ColumnCount);
+    QStringList headers{i18n("DataSource/Key"), i18n("Type"), i18n("Value")};
     m_dataModel->setHorizontalHeaderLabels(headers);
+    m_data->setColumnWidth(ColumnDataSourceAndKey, 200);
     m_engine = nullptr;
     m_sourceCount = 0;
 
@@ -192,7 +204,8 @@ void EngineExplorer::showEngine(const QString &name)
     // qDebug() << "we have " << sources.count() << " data sources";
     connect(m_engine, &Plasma::DataEngine::sourceAdded, this, &EngineExplorer::addSource);
     connect(m_engine, &Plasma::DataEngine::sourceRemoved, this, &EngineExplorer::removeSource);
-    foreach (const QString &source, m_engine->sources()) {
+    const QStringList &sources = m_engine->sources();
+    for (const QString &source : sources) {
         // qDebug() << "adding " << source;
         addSource(source);
     }
@@ -216,6 +229,7 @@ void EngineExplorer::addSource(const QString &source)
     }
 
     QStandardItem *parent = new QStandardItem(source);
+    parent->setToolTip(source);
     m_dataModel->appendRow(parent);
 
     // qDebug() << "getting data for source " << source;
@@ -232,13 +246,13 @@ void EngineExplorer::addSource(const QString &source)
 
 void EngineExplorer::removeSource(const QString &source)
 {
-    QList<QStandardItem *> items = m_dataModel->findItems(source, Qt::MatchExactly);
+    const QList<QStandardItem *> items = m_dataModel->findItems(source, Qt::MatchExactly);
 
     if (items.count() < 1) {
         return;
     }
 
-    foreach (QStandardItem *item, items) {
+    for (QStandardItem *item : items) {
         m_dataModel->removeRow(item->row());
     }
 
@@ -274,7 +288,7 @@ void EngineExplorer::showDataContextMenu(const QPoint &point)
 {
     QModelIndex index = m_data->indexAt(point);
     if (index.isValid()) {
-        if (index.parent().isValid()) {
+        while (index.parent().isValid()) {
             index = index.parent();
         }
 
@@ -343,9 +357,7 @@ QString EngineExplorer::convertToString(const QVariant &value)
         QVariantMap map = value.toMap();
         QString str = i18np("<1 item>", "<%1 items>", map.size());
 
-        QMapIterator<QString, QVariant> it(map);
-        while (it.hasNext()) {
-            it.next();
+        for (auto it = map.constBegin(); it != map.constEnd(); it++) {
             str += "\n" + it.key() + ": " + convertToString(it.value());
         }
 
@@ -371,9 +383,14 @@ QString EngineExplorer::convertToString(const QVariant &value)
         QRectF rectF = value.toRectF();
         return QString("<x:%1, y:%2, w:%3, h:%4>").arg(rectF.x()).arg(rectF.y()).arg(rectF.width()).arg(rectF.height());
     }
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
     case QVariant::RegExp: {
         return QString("%1").arg(value.toRegExp().pattern());
     }
+#endif
+    case QVariant::RegularExpression: {
+        return value.toRegularExpression().pattern();
+    }
     case QVariant::Region: {
         QRect region = value.value<QRegion>().boundingRect();
         return QString("<x:%1, y:%2, w:%3, h:%4>").arg(region.x()).arg(region.y()).arg(region.width()).arg(region.height());
@@ -409,19 +426,18 @@ QString EngineExplorer::convertToString(const QVariant &value)
         Plasma::DataEngine::Data data = value.value<Plasma::DataEngine::Data>();
         if (!data.isEmpty()) {
             QStringList result;
-            QMapIterator<QString, QVariant> it(data);
 
-            while (it.hasNext()) {
-                it.next();
+            for (auto it = data.constBegin(); it != data.constEnd(); it++) {
                 result << (it.key() + ": " + convertToString(it.value()));
             }
 
             return result.join("\n");
-        } else if (value.canConvert(QVariant::String)) {
-            if (value.toString().isEmpty()) {
+        } else if (value.canConvert<QString>()) {
+            QString str = value.toString();
+            if (str.isEmpty()) {
                 return i18nc("The user did a query to a dataengine and it returned empty data", "<empty>");
             } else {
-                return value.toString();
+                return str;
             }
         }
 
@@ -433,41 +449,88 @@ QString EngineExplorer::convertToString(const QVariant &value)
 int EngineExplorer::showData(QStandardItem *parent, Plasma::DataEngine::Data data)
 {
     int rowCount = 0;
-    Plasma::DataEngine::DataIterator it(data);
-    // parent->insertRows(0, data.count());
-    // parent->setColumnCount(3);
-    while (it.hasNext()) {
-        it.next();
-        parent->setChild(rowCount, 1, new QStandardItem(it.key()));
-        if (it.value().canConvert(QVariant::List) /* && ! it.value().type() == QVariant::StringList
-                                                     */) {
-            bool first = true;
-            foreach (const QVariant &var, it.value().toList()) {
-                QStandardItem *item = new QStandardItem(convertToString(var));
-                if (!first) {
-                    parent->setChild(rowCount, 1, new QStandardItem(QString()));
-                }
-                item->setToolTip(item->text());
-                parent->setChild(rowCount, 2, item);
-                parent->setChild(rowCount, 3, new QStandardItem(var.typeName()));
-                first = false;
-                ++rowCount;
-            }
-        } else {
-            QStandardItem *item;
-            if (it.value().canConvert<QIcon>()) {
-                item = new QStandardItem(it.value().value<QIcon>(), "");
-            } else {
-                item = new QStandardItem(convertToString(it.value()));
-            }
-
-            item->setToolTip(item->text());
-            parent->setChild(rowCount, 2, item);
-            parent->setChild(rowCount, 3, new QStandardItem(it.value().typeName()));
-            ++rowCount;
-        }
+    for (auto it = data.constBegin(); it != data.constEnd(); it++) {
+        showData(parent, rowCount++, it.key(), it.value());
     }
+    return rowCount;
+}
 
+void EngineExplorer::showData(QStandardItem *parent, int row, const QString &key, const QVariant &value)
+{
+    static_assert(ColumnDataSourceAndKey == 0);
+    // QTreeView only expands tree for children of column #zero.
+    QStandardItem *current = new QStandardItem(key);
+    current->setToolTip(key);
+    parent->setChild(row, ColumnDataSourceAndKey, current);
+
+    const char *typeName = value.typeName();
+    int rowCount = 0;
+
+    if (value.userType() == qMetaTypeId<QList<QVariantMap>>()) {
+        // this case is a bit special, and has to be handled before generic QVariantList
+        const QList<QVariantMap> list = value.value<QList<QVariantMap>>();
+        rowCount = showContainerData(parent, current, row, typeName, list);
+    } else if (value.canConvert<QVariantList>()) {
+        const QVariantList list = value.toList();
+        rowCount = showContainerData(parent, current, row, typeName, list);
+    } else if (value.canConvert<QVariantMap>()) {
+        const QVariantMap map = value.toMap();
+        rowCount = showContainerData(parent, current, row, typeName, map);
+    } else {
+        parent->setChild(row, ColumnType, new QStandardItem(typeName));
+        // clang-format off
+        QStandardItem *item = value.canConvert<QIcon>()
+            ? new QStandardItem(value.value<QIcon>(), "")
+            : new QStandardItem(convertToString(value));
+        // clang-format on
+        item->setToolTip(item->text());
+        parent->setChild(row, ColumnValue, item);
+        // leave rowCount at value 0
+    }
+    removeExtraRows(current, rowCount);
+}
+
+int EngineExplorer::showContainerData(QStandardItem *parent, QStandardItem *current, int row, const char *typeName, const QList<QVariantMap> &list)
+{
+    QStandardItem *typeItem = new QStandardItem(typeName);
+    typeItem->setToolTip(typeItem->text());
+    parent->setChild(row, ColumnType, typeItem);
+    parent->setChild(row, ColumnValue, new QStandardItem(ki18ncp("Length of the list", "<%1 item>", "<%1 items>").subs(list.length()).toString()));
+
+    int rowCount = 0;
+    for (const QVariantMap &map : list) {
+        showData(current, rowCount, QString::number(rowCount), map);
+        rowCount++;
+    }
+    return rowCount;
+}
+
+int EngineExplorer::showContainerData(QStandardItem *parent, QStandardItem *current, int row, const char *typeName, const QVariantList &list)
+{
+    QStandardItem *typeItem = new QStandardItem(typeName);
+    typeItem->setToolTip(typeItem->text());
+    parent->setChild(row, ColumnType, typeItem);
+    parent->setChild(row, ColumnValue, new QStandardItem(ki18ncp("Length of the list", "<%1 item>", "<%1 items>").subs(list.length()).toString()));
+
+    int rowCount = 0;
+    for (const QVariant &var : list) {
+        showData(current, rowCount, QString::number(rowCount), var);
+        rowCount++;
+    }
+    return rowCount;
+}
+
+int EngineExplorer::showContainerData(QStandardItem *parent, QStandardItem *current, int row, const char *typeName, const QVariantMap &map)
+{
+    QStandardItem *typeItem = new QStandardItem(typeName);
+    typeItem->setToolTip(typeItem->text());
+    parent->setChild(row, ColumnType, typeItem);
+    parent->setChild(row, ColumnValue, new QStandardItem(ki18ncp("Size of the map", "<%1 pair>", "<%1 pairs>").subs(map.size()).toString()));
+
+    int rowCount = 0;
+    for (auto it = map.constBegin(); it != map.constEnd(); it++) {
+        showData(current, rowCount++, it.key(), it.value());
+    }
     return rowCount;
 }
 
