@@ -34,16 +34,19 @@ void Delegate::paint(QPainter *painter, const QStyleOptionViewItem &option, cons
         return;
     }
 
+    const QHash<int, QByteArray> roleNames = index.model()->roleNames();
+    const QList<int> roles = roleNames.keys();
+
     QFontMetrics fm(option.font);
     int maxWidth = 0;
-    foreach (int role, index.model()->roleNames().keys()) {
-        const QString text = index.model()->roleNames().value(role) + QLatin1String(": ");
+    for (int role : roles) {
+        const QString text = roleNames.value(role) + QLatin1String(": ");
         maxWidth = qMax(maxWidth, fm.boundingRect(text).width());
     }
 
     int i = 2;
-    foreach (int role, index.model()->roleNames().keys()) {
-        const QString text = index.model()->roleNames().value(role) + QLatin1String(": ");
+    for (int role : roles) {
+        const QString text = roleNames.value(role) + QLatin1String(": ");
         painter->drawText(option.rect.x() + maxWidth - fm.boundingRect(text).width(), option.rect.y() + i * fm.height(), text);
 
         if (index.data(role).canConvert<QIcon>()) {
@@ -82,10 +85,10 @@ ModelViewer::ModelViewer(Plasma::DataEngine *engine, const QString &source, QWid
     QString engineName = i18nc("Plasma engine with unknown name", "Unknown");
 
     if (m_engine) {
-        if (m_engine->pluginInfo().isValid()) {
-            engineName = KStringHandler::capwords(m_engine->pluginInfo().name());
+        if (m_engine->metadata().isValid()) {
+            engineName = KStringHandler::capwords(m_engine->metadata().name());
         }
-        qDebug() << "########### CALLING SERVICE FOR SOURCE: " << m_source;
+        qDebug() << "Requesting model for source:" << m_source;
         m_model = m_engine->modelForSource(m_source);
 
         if (m_model != nullptr) {
